#!/usr/bin/env python
####################################################################################################
# NAME
#    <NAME> - contain graphical utility functions for charts
#
# SYNOPSIS
#    <NAME>
#
# AUTHOR
#    Written by Florian Barras (florian@barras.io).
#
# COPYRIGHT
#    Copyright © 2013-2022 Florian Barras <https://barras.io>.
#    The MIT License (MIT) <https://opensource.org/licenses/MIT>.
####################################################################################################

import io

import matplotlib.figure as mfigure
import matplotlib.pyplot as mplot
import matplotlib.ticker as mticker
import plotly.express as px
import plotly.graph_objs as go
import plotly.io as pio
import plotly.subplots as sp
import plotly.tools as tls

import ngui.web as web
from ngui.image import *
from nutil.math import *

####################################################################################################
# CHART SETTINGS
####################################################################################################

pio.renderers.default = 'browser'

####################################################################################################
# CHART CONSTANTS
####################################################################################################

__CHART_CONSTANTS___________________________________ = ''

# The default tick length
DEFAULT_TICK_LENGTH = 4

# The default tick direction
DEFAULT_TICK_DIRECTION = 'outside'

##################################################

MAP_PROJECTIONS = [
	'equirectangular', 'mercator', 'orthographic', 'natural earth', 'kavrayskiy7', 'miller',
	'robinson', 'eckert4', 'azimuthal equal area', 'azimuthal equidistant', 'conic equal area',
	'conic conformal', 'conic equidistant', 'gnomonic', 'stereographic', 'mollweide', 'hammer',
	'transverse mercator', 'albers usa', 'winkel tripel', 'aitoff', 'sinusoidal'
]

####################################################################################################
# CHART FUNCTIONS
####################################################################################################

__CHART___________________________________________ = ''


def is_matplot(fig):
	return isinstance(fig, mfigure.Figure)


def is_plotly(fig):
	return isinstance(fig, go._figure.Figure)


#########################

def is_multi_plot(fig):
	return not is_null(fig._grid_ref)


##################################################

def get_grid_size(n, row_count=None, col_count=None):
	if is_collection(n):
		n = max(1, count_cols(n))
	if is_null(col_count):
		col_count = round_to_int(sqrt(n))
	if is_null(row_count):
		row_count = ceil(n / col_count)
	return row_count, col_count


def get_hover_template(index):
	if not is_empty(index):
		return collapse('<b>%{customdata}</b><br />',
		                '<b>x:</b> %{x}<br />',
		                '<b>y:</b> %{y}')
	return collapse('<b>x:</b> %{x}<br />',
	                '<b>y:</b> %{y}')


def get_label(data, transformation=None, yaxis=0,
              show_date=False, show_name=True):
	if is_null(data):
		return ''
	yaxis = '(' + str(yaxis) + ')' if yaxis != 0 else ''
	if show_date and is_time_series(data):
		date_from = get_first(data.index)
		date_to = get_last(data.index)
		year_from = date_from.year if not is_null(date_from) else None
		year_to = date_to.year if not is_null(date_to) else None
		if is_any_null(year_from, year_to):
			date_range = ''
		elif year_from != year_to:
			date_range = collapse(year_from, '-', year_to)
		else:
			date_range = year_from
	else:
		date_range = ''
	if show_name and not is_empty(data):
		name = get_names(data)[0] if is_collection(data) else data
		name = str(name).title() if not is_null(name) else ''
	else:
		name = ''
	transformation = transformation.value.title() if not is_null(transformation) else ''
	return paste(yaxis, date_range, name, transformation)


def get_margin(x, fig=None, has_title=False, has_title_x=False, has_title_y=False):
	'''Returns the margin with the specified ratio to the width or height.'''
	if not is_null(fig) and not is_empty(fig._layout_obj.annotations):
		has_title |= (not is_empty(fig._layout_obj.title.text) or
		              any(a.yanchor in ('bottom', 'top') for a in fig._layout_obj.annotations))
		has_title_x |= (not is_empty(fig._layout_obj.xaxis.title.text) or
		                any(a.yanchor in ('bottom', 'top') for a in fig._layout_obj.annotations))
		has_title_y |= (not is_empty(fig._layout_obj.yaxis.title.text) or
		                any(a.xanchor in ('left', 'right') for a in fig._layout_obj.annotations))
	if is_null(x):
		x = {}
	if is_dict(x):
		for k in ('l', 'r', 'b', 't'):
			if k not in x:
				x[k] = (DEFAULT_MARGIN_WITH_TITLE[k] if (has_title and k == 't' or
				                                         has_title_x and k == 'b' or
				                                         has_title_y and k in ('l', 'r')) else
				        DEFAULT_MARGIN[k])
		return x
	return dict(l=x, r=x, b=x, t=x)


##################################################

def matplot_to_plotly(fig, resize=False, strip_style=False, verbose=VERBOSE):
	for ax in fig.axes:
		ax.xaxis._gridOnMajor = ax.xaxis._major_tick_kw['gridOn']
		ax.yaxis._gridOnMajor = ax.yaxis._major_tick_kw['gridOn']
		for collection in ax.collections:
			collection.get_offset_position = lambda: 'screen'
		for _, spine in ax.spines.items():
			spine.is_frame_like = lambda: False
	return tls.mpl_to_plotly(fig, resize=resize, strip_style=strip_style, verbose=verbose)


#########################

def fig_to_image(fig, format,
                 scale=DEFAULT_SCALE,
                 width=DEFAULT_WIDTH, height=DEFAULT_HEIGHT, margin=None):
	'''Converts the specified figure to an image buffer with the specified format.'''
	update_layout_size(fig, width=width, height=height, margin=margin)
	if is_matplot(fig):
		buffer = io.BytesIO()
		fig.savefig(buffer, format=format, dpi=scale * 100)
		buffer.seek(0)
		return buffer.read()
	elif is_plotly(fig):
		return pio.to_image(fig, format=format, scale=scale, width=width, height=height)


def fig_to_jpg(fig,
               scale=DEFAULT_SCALE,
               width=DEFAULT_WIDTH, height=DEFAULT_HEIGHT, margin=None):
	'''Converts the specified figure to a JPEG image.'''
	return fig_to_image(fig, JPEG_FORMAT, scale=scale, width=width, height=height, margin=margin)


def fig_to_png(fig,
               scale=DEFAULT_SCALE,
               width=DEFAULT_WIDTH, height=DEFAULT_HEIGHT, margin=None):
	'''Converts the specified figure to a PNG image.'''
	return fig_to_image(fig, PNG_FORMAT, scale=scale, width=width, height=height, margin=margin)


def fig_to_svg(fig,
               scale=DEFAULT_SCALE,
               width=DEFAULT_WIDTH, height=DEFAULT_HEIGHT, margin=None):
	'''Converts the specified figure to an SVG image.'''
	return fig_to_image(fig, SVG_FORMAT, scale=scale, width=width, height=height, margin=margin)


def fig_to_webp(fig,
                scale=DEFAULT_SCALE,
                width=DEFAULT_WIDTH, height=DEFAULT_HEIGHT, margin=None):
	'''Converts the specified figure to a WebP image.'''
	return fig_to_image(fig, WEBP_FORMAT, scale=scale, width=width, height=height, margin=margin)


#########################

def fig_to_html(fig, full=True,
                width=DEFAULT_WIDTH, height=DEFAULT_HEIGHT, margin=None):
	'''Converts the specified figure to HTML.'''
	if is_null(fig):
		return ''
	update_layout_size(fig, width=width, height=height, margin=margin)
	return pio.to_html(fig, full_html=full, default_width=width, default_height=height)


def fig_to_image_html(fig, format, encoding=DEFAULT_ENCODING, mode=DEFAULT_IMAGE_MODE, style=None,
                      rotate=False, scale=DEFAULT_SCALE,
                      width=DEFAULT_WIDTH, height=DEFAULT_HEIGHT, margin=None):
	'''Converts the specified figure to the specified format, encodes it to Base64 and returns its
	HTML code.'''
	if is_null(fig):
		return ''
	image = fig_to_image(fig, format, scale=scale, width=width, height=height, margin=margin)
	return image_to_html(image, format, encoding=encoding, mode=mode, style=style,
	                     rotate=rotate,
	                     width=width, height=height)


def fig_to_jpg_html(fig, encoding=DEFAULT_ENCODING, style=None,
                    rotate=False, scale=DEFAULT_SCALE,
                    width=DEFAULT_WIDTH, height=DEFAULT_HEIGHT, margin=None):
	'''Converts the specified figure to a JPEG image, encodes it to Base64 and returns its HTML
	code.'''
	return fig_to_image_html(fig, JPEG_FORMAT, encoding=encoding, style=style,
	                         rotate=rotate, scale=scale,
	                         width=width, height=height, margin=margin)


def fig_to_png_html(fig, encoding=DEFAULT_ENCODING, style=None,
                    rotate=False, scale=DEFAULT_SCALE,
                    width=DEFAULT_WIDTH, height=DEFAULT_HEIGHT, margin=None):
	'''Converts the specified figure to a PNG image, encodes it to Base64 and returns its HTML
	code.'''
	return fig_to_image_html(fig, PNG_FORMAT, encoding=encoding, style=style,
	                         rotate=rotate, scale=scale,
	                         width=width, height=height, margin=margin)


def fig_to_svg_html(fig, encoding=DEFAULT_ENCODING, style=None,
                    rotate=False, scale=DEFAULT_SCALE,
                    width=DEFAULT_WIDTH, height=DEFAULT_HEIGHT, margin=None):
	'''Converts the specified figure to an SVG image, encodes it to Base64 and returns its HTML
	code.'''
	return fig_to_image_html(fig, SVG_FORMAT, encoding=encoding, style=style,
	                         rotate=rotate, scale=scale,
	                         width=width, height=height, margin=margin)


def fig_to_webp_html(fig, encoding=DEFAULT_ENCODING, style=None,
                     rotate=False, scale=DEFAULT_SCALE,
                     width=DEFAULT_WIDTH, height=DEFAULT_HEIGHT, margin=None):
	'''Converts the specified figure to a WEBP image, encodes it to Base64 and returns its HTML
	code.'''
	return fig_to_image_html(fig, WEBP_FORMAT, encoding=encoding, style=style,
	                         rotate=rotate, scale=scale,
	                         width=width, height=height, margin=margin)


##################################################

def create_figure(auto_size=True,
                  axis_color='black', axis_width=DEFAULT_LINE_WIDTH,
                  bar_mode=None,
                  bg_color=DEFAULT_BG_COLOR,
                  font_size=DEFAULT_FONT_SIZE,
                  grid_color='lightgray', grid_width=1,
                  label_color='black', label_size=None,
                  legend_bg_color=DEFAULT_BG_COLOR, legend_x=0.01, legend_y=0.99,
                  range_to_zero_x=False, range_to_zero_y=False, range_to_zero_y2=False,
                  show_grid_x=True, show_grid_y=True, show_grid_y2=True, show_spine=True,
                  show_title=True, show_zero_line=True,
                  tick_color='black', tick_direction=DEFAULT_TICK_DIRECTION,
                  tick_length=DEFAULT_TICK_LENGTH,
                  tick_number_x=None, tick_number_y=None, tick_number_y2=None,
                  tick_start_x=None, tick_start_y=None, tick_start_y2=None,
                  tick_step_x=None, tick_step_y=None, tick_step_y2=None,
                  tick_values_x=None, tick_values_y=None, tick_values_y2=None,
                  title=None, title_x=None, title_y=None, title_y2=None,
                  width=DEFAULT_WIDTH, height=DEFAULT_HEIGHT, margin=None,
                  zero_line_color='darkgray', zero_line_width=DEFAULT_LINE_WIDTH):
	fig = go.Figure()
	update_layout(fig,
	              auto_size=auto_size,
	              axis_color=axis_color, axis_width=axis_width,
	              bar_mode=bar_mode,
	              bg_color=bg_color,
	              font_size=font_size,
	              grid_color=grid_color, grid_width=grid_width,
	              label_color=label_color, label_size=label_size,
	              legend_bg_color=legend_bg_color, legend_x=legend_x, legend_y=legend_y,
	              range_to_zero_x=range_to_zero_x, range_to_zero_y=range_to_zero_y,
	              range_to_zero_y2=range_to_zero_y2,
	              show_grid_x=show_grid_x, show_grid_y=show_grid_y, show_grid_y2=show_grid_y2,
	              show_spine=show_spine, show_title=show_title, show_zero_line=show_zero_line,
	              tick_color=tick_color, tick_direction=tick_direction, tick_length=tick_length,
	              tick_number_x=tick_number_x, tick_number_y=tick_number_y,
	              tick_number_y2=tick_number_y2,
	              tick_start_x=tick_start_x, tick_start_y=tick_start_y, tick_start_y2=tick_start_y2,
	              tick_step_x=tick_step_x, tick_step_y=tick_step_y, tick_step_y2=tick_step_y2,
	              tick_values_x=tick_values_x, tick_values_y=tick_values_y,
	              tick_values_y2=tick_values_y2,
	              title=title, title_x=title_x, title_y=title_y, title_y2=title_y2,
	              width=width, height=height, margin=margin,
	              zero_line_color=zero_line_color, zero_line_width=zero_line_width)
	return fig


def create_figures(row_count, col_count,
                   share_x=False, share_y=False,
                   auto_size=True,
                   axis_color='black', axis_width=DEFAULT_LINE_WIDTH,
                   bar_mode=None,
                   bg_color=DEFAULT_BG_COLOR,
                   font_size=DEFAULT_FONT_SIZE,
                   grid_color='lightgray', grid_width=1,
                   label_color='black', label_size=None,
                   legend_bg_color=DEFAULT_BG_COLOR, legend_x=0.01, legend_y=0.99,
                   range_to_zero_x=False, range_to_zero_y=False,
                   show_grid_x=True, show_grid_y=True, show_spine=True,
                   show_title=True, show_zero_line=True,
                   tick_color='black', tick_direction=DEFAULT_TICK_DIRECTION,
                   tick_length=DEFAULT_TICK_LENGTH,
                   tick_number_x=None, tick_number_y=None,
                   tick_start_x=None, tick_start_y=None,
                   tick_step_x=None, tick_step_y=None,
                   tick_values_x=None, tick_values_y=None,
                   title=None, subtitles=None, title_x=None, title_y=None,
                   width=DEFAULT_WIDTH, height=DEFAULT_HEIGHT, margin=None,
                   zero_line_color='darkgray', zero_line_width=DEFAULT_LINE_WIDTH):
	margin = get_margin(margin, has_title=not is_empty(title) or not is_empty(subtitles),
	                    has_title_x=not is_empty(title_x), has_title_y=not is_empty(title_y))
	fig = sp.make_subplots(rows=row_count, cols=col_count,
	                       shared_xaxes=share_x, shared_yaxes=share_y,
	                       subplot_titles=apply(subtitles, to_string),
	                       x_title=title_x, y_title=title_y)
	update_layout(fig,
	              auto_size=auto_size,
	              axis_color=axis_color, axis_width=axis_width,
	              bar_mode=bar_mode,
	              bg_color=bg_color,
	              font_size=font_size,
	              grid_color=grid_color, grid_width=grid_width,
	              label_color=label_color, label_size=label_size,
	              legend_bg_color=legend_bg_color, legend_x=legend_x, legend_y=legend_y,
	              range_to_zero_x=range_to_zero_x, range_to_zero_y=range_to_zero_y,
	              show_grid_x=show_grid_x, show_grid_y=show_grid_y,
	              show_spine=show_spine, show_title=show_title, show_zero_line=show_zero_line,
	              tick_color=tick_color, tick_direction=tick_direction, tick_length=tick_length,
	              tick_number_x=tick_number_x, tick_number_y=tick_number_y,
	              tick_start_x=tick_start_x, tick_start_y=tick_start_y,
	              tick_step_x=tick_step_x, tick_step_y=tick_step_y,
	              tick_values_x=tick_values_x, tick_values_y=tick_values_y,
	              title=title, title_x=None, title_y=None,
	              width=width, height=height, margin=margin,
	              zero_line_color=zero_line_color, zero_line_width=zero_line_width)
	return fig


#########################

def create_choropleth_map(df, loc_col, label_col, loc_mode='ISO-3', label_name=None,
                          # Layout
                          title=None, dragmode=False, showframe=False,
                          colors=get_RYG(), range_color=None,
                          width=DEFAULT_WIDTH, height=DEFAULT_HEIGHT, margin=None,
                          # Geos
                          lat=None, lon=None,
                          projection='miller',
                          range_mode='auto', lataxis_range=None, lonaxis_range=None,
                          resolution=50,
                          showcoastlines=True, coastlinecolor='Black',
                          showland=False, landcolor='LightGreen',
                          showocean=True, oceancolor='AliceBlue',
                          showlakes=False, lakecolor='Blue',
                          showrivers=False, rivercolor='Blue'):
	'''Creates a choropleth map with the specified parameters.'''
	fig = px.choropleth(data_frame=df,
	                    lat=lat, lon=lon,
	                    locations=loc_col, locationmode=loc_mode, projection=projection,
	                    labels={label_col: label_name if not is_null(label_name) else label_col},
	                    color=label_col, range_color=range_color,
	                    color_continuous_scale=colors, color_discrete_sequence=colors)
	update_layout(fig, title=title, width=width, height=height, margin=margin)
	fig.update_layout(clickmode='event+select', dragmode=dragmode, hovermode='closest')
	if range_mode == 'auto':
		if is_null(projection) or projection in ('equirectangular', 'kavrayskiy7', 'sinusoidal'):
			lataxis_range = [-48, 63]
		elif projection == 'aitoff':
			lataxis_range = [-39, 63]
		elif projection == 'eckert4':
			lataxis_range = [-50, 59]
		elif projection == 'hammer':
			lataxis_range = [-42, 61]
		elif projection == 'mercator':
			lataxis_range = [-35, 71]
		elif projection == 'miller':
			lataxis_range = [-43, 68]
		elif projection == 'mollweide':
			lataxis_range = [-50, 61]
		elif projection == 'natural earth' or projection == 'robinson':
			lataxis_range = [-49, 62]
		elif projection == 'winkel tripel':
			lataxis_range = [-44, 64]
	fig.update_geos(
		lataxis_range=lataxis_range,
		lonaxis_range=lonaxis_range,
		resolution=resolution,
		showframe=showframe,
		showcoastlines=showcoastlines, coastlinecolor=coastlinecolor,
		showland=showland, landcolor=landcolor,
		showocean=showocean, oceancolor=oceancolor,
		showlakes=showlakes, lakecolor=lakecolor,
		showrivers=showrivers, rivercolor=rivercolor)
	return fig


##################################################

def draw(x, y=None,
         # Chart
         color=None, dash=None, fill='none', index=None, line_width=DEFAULT_LINE_WIDTH,
         marker_size=DEFAULT_MARKER_SIZE, mode='lines', name=None, opacity=1, stackgroup=None,
         yaxis=0,
         # Flags
         show_date=False, show_legend=True, show_name=True):
	data = x
	if is_null(y):
		x = to_array(get_index(data))
		y = get_values(data)
	elif is_null(index) and is_frame(data):
		index = to_array(get_index(data))
	if is_null(name):
		name = get_name(data)
	name = get_label(name, yaxis=yaxis, show_date=show_date, show_name=show_name)
	hover_template = get_hover_template(index)
	line = dict(color=color, dash=dash, width=line_width)
	marker = dict(color=color, size=marker_size)
	if mode == 'lines':
		marker = None
	elif mode == 'markers':
		line = None
	return go.Scatter(x=x, y=y,
	                  # Chart
	                  customdata=index, hovertemplate=hover_template, fill=fill, line=line,
	                  marker=marker, mode=mode, name=name, opacity=opacity, stackgroup=stackgroup,
	                  yaxis='y' + str(1 if yaxis == 0 else yaxis),
	                  # Flags
	                  showlegend=show_legend)


def draw_ellipse(center, a, b, angle=0, precision=100,
                 # Chart
                 color=None, dash=None, fill='none', index=None, line_width=DEFAULT_LINE_WIDTH,
                 marker_size=DEFAULT_MARKER_SIZE, mode='lines', name=None, opacity=1, yaxis=0,
                 # Flags
                 show_date=False, show_legend=True, show_name=True):
	X, Y = create_ellipse(center, a, b, angle=angle, precision=precision)
	return draw(x=X, y=Y,
	            # Chart
	            color=color, dash=dash, fill=fill, index=index, line_width=line_width,
	            marker_size=marker_size, mode=mode, name=name, opacity=opacity, yaxis=yaxis,
	            # Flags
	            show_date=show_date, show_legend=show_legend, show_name=show_name)


def draw_series(series, *args, f=None,
                # Chart
                color=None, dash=None, fill='none', index=None, line_width=DEFAULT_LINE_WIDTH,
                marker_size=DEFAULT_MARKER_SIZE, mode='lines', name=None, opacity=1,
                stackgroup=None, yaxis=0,
                # Flags
                show_date=False, show_legend=True, show_name=True, **kwargs):
	if not is_null(f):
		series = f(series, *args, **kwargs)
	return draw(series,
	            # Chart
	            color=color, dash=dash, fill=fill, index=index, line_width=line_width,
	            marker_size=marker_size, mode=mode, name=name, opacity=opacity,
	            stackgroup=stackgroup, yaxis=yaxis,
	            # Flags
	            show_date=show_date, show_legend=show_legend, show_name=show_name)


#########################

def plot_multi(df, draw, *args,
               # Figure
               fig=None, row_count=None, col_count=None,
               share_x=True, share_y=True,
               title=None, subtitles=None, title_x=None, title_y=None,
               width=DEFAULT_WIDTH, height=DEFAULT_HEIGHT, margin=None,
               # Chart
               colors=DEFAULT_COLORS, **kwargs):
	row_count, col_count = get_grid_size(df, row_count=row_count, col_count=col_count)
	if is_null(fig):
		if is_null(subtitles):
			subtitles = get_names(df)
		fig = create_figures(row_count, col_count,
		                     share_x=share_x, share_y=share_y,
		                     title=title, subtitles=subtitles, title_x=title_x, title_y=title_y,
		                     width=width, height=height, margin=margin)
	colors = get_iterator(to_list(colors), cycle=True)

	# Get the number of series
	series_count = count_cols(df)

	for i in range(row_count):
		for j in range(col_count):
			series_index = i * col_count + j
			if series_index < series_count:
				fig.add_trace(draw(get_col(df, series_index), *args, color=next(colors), **kwargs),
				              row=i + 1, col=j + 1)
	return fig


def plot_series(series, *args, f=None,
                # Figure
                fig=None, title=None, title_x=None, title_y=None, title_y2=None,
                width=DEFAULT_WIDTH, height=DEFAULT_HEIGHT, margin=None,
                # Chart
                colors=DEFAULT_COLORS, dash=None, fill='none', index=None,
                line_width=DEFAULT_LINE_WIDTH, marker_size=DEFAULT_MARKER_SIZE, mode='lines',
                name=None, opacity=1, stackgroup=None, yaxis=0,
                # Flags
                show_date=False, show_legend=True, show_name=True, **kwargs):
	if is_null(fig):
		fig = create_figure(title=title, title_x=title_x, title_y=title_y, title_y2=title_y2,
		                    width=width, height=height, margin=margin)
	colors = get_iterator(to_list(colors), cycle=True)

	for s in to_series(series) if is_frame(series) else [series]:
		fig.add_trace(draw_series(s,
		                          *args, f=f,
		                          # Chart
		                          color=next(colors), dash=dash, fill=fill, index=index,
		                          line_width=line_width, marker_size=marker_size, mode=mode,
		                          name=name, opacity=opacity, stackgroup=stackgroup, yaxis=yaxis,
		                          # Flags
		                          show_date=show_date, show_legend=show_legend, show_name=show_name,
		                          **kwargs))
	return fig


def plot_multi_series(df, *args, f=None,
                      # Figure
                      fig=None, row_count=None, col_count=None, share_x=True, share_y=True,
                      title=None, subtitles=None, title_x=None, title_y=None,
                      width=DEFAULT_WIDTH, height=DEFAULT_HEIGHT, margin=None,
                      # Chart
                      colors=DEFAULT_COLORS, dash=None, fill='none', index=None,
                      line_width=DEFAULT_LINE_WIDTH, marker_size=DEFAULT_MARKER_SIZE, mode='lines',
                      opacity=1, stackgroup=None,
                      # Flags
                      show_date=False, show_legend=False, show_name=True, **kwargs):
	return plot_multi(df, draw_series,
	                  *args, f=f,
	                  # Figure
	                  fig=fig, row_count=row_count, col_count=col_count,
	                  share_x=share_x, share_y=share_y,
	                  title=title, subtitles=subtitles, title_x=title_x, title_y=title_y,
	                  width=width, height=height, margin=margin,
	                  # Chart
	                  colors=colors, dash=dash, fill=fill, index=index, line_width=line_width,
	                  marker_size=marker_size, mode=mode, opacity=opacity, stackgroup=stackgroup,
	                  # Flags
	                  show_date=show_date, show_legend=show_legend, show_name=show_name, **kwargs)


def plot_ellipse(center, a, b, angle=0, precision=100,
                 # Figure
                 fig=None, title=None, title_x=None, title_y=None, title_y2=None,
                 width=DEFAULT_WIDTH, height=DEFAULT_HEIGHT, margin=None,
                 # Chart
                 color=None, dash=None, fill='none', index=None, line_width=DEFAULT_LINE_WIDTH,
                 marker_size=DEFAULT_MARKER_SIZE, mode='lines', name=None, opacity=1, yaxis=0,
                 # Flags
                 show_axes=True, show_date=False, show_legend=True, show_name=True):
	if is_null(fig):
		fig = create_figure(title=title, title_x=title_x, title_y=title_y, title_y2=title_y2,
		                    width=width, height=height, margin=margin)
	fig.add_trace(draw_ellipse(center, a, b, angle=angle, precision=precision,
	                           # Chart
	                           color=color, dash=dash, fill=fill, index=index,
	                           line_width=line_width, marker_size=marker_size, mode=mode, name=name,
	                           opacity=opacity, yaxis=yaxis,
	                           # Flags
	                           show_date=show_date, show_legend=show_legend, show_name=show_name))
	if show_axes:
		fig.add_trace(draw([center[0] - a * cos(angle), center[0] + a * cos(angle)],
		                   [center[1] - a * sin(angle), center[1] + a * sin(angle)],
		                   # Chart
		                   color=color, dash=dash, line_width=line_width, marker_size=marker_size,
		                   name=name, opacity=opacity,
		                   # Flags
		                   show_legend=False))
		fig.add_trace(draw([center[0] - b * sin(angle), center[0] + b * sin(angle)],
		                   [center[1] + b * cos(angle), center[1] - b * cos(angle)],
		                   # Chart
		                   color=color, dash=dash, line_width=line_width, marker_size=marker_size,
		                   name=name, opacity=opacity,
		                   # Flags
		                   show_legend=False))
	return fig


#########################

def update_layout(fig,
                  auto_size=True,
                  axis_color='black', axis_width=DEFAULT_LINE_WIDTH,
                  bar_mode=None,
                  bg_color=DEFAULT_BG_COLOR,
                  font_size=DEFAULT_FONT_SIZE,
                  grid_color='lightgray', grid_width=1,
                  label_color='black', label_size=None,
                  legend_bg_color=DEFAULT_BG_COLOR, legend_x=0.01, legend_y=0.99,
                  range_to_zero_x=False, range_to_zero_y=False, range_to_zero_y2=False,
                  show_grid_x=True, show_grid_y=True, show_grid_y2=True, show_spine=True,
                  show_title=True, show_zero_line=True,
                  tick_color='black', tick_direction=DEFAULT_TICK_DIRECTION,
                  tick_length=DEFAULT_TICK_LENGTH,
                  tick_number_x=None, tick_number_y=None, tick_number_y2=None,
                  tick_start_x=None, tick_start_y=None, tick_start_y2=None,
                  tick_step_x=None, tick_step_y=None, tick_step_y2=None,
                  tick_values_x=None, tick_values_y=None, tick_values_y2=None,
                  title=None, title_x=None, title_y=None, title_y2=None,
                  width=DEFAULT_WIDTH, height=DEFAULT_HEIGHT, margin=None,
                  zero_line_color='darkgray', zero_line_width=DEFAULT_LINE_WIDTH):
	margin = get_margin(margin, fig=fig, has_title=not is_empty(title),
	                    has_title_x=not is_empty(title_x), has_title_y=not is_empty(title_y))
	update_layout_plot(fig, auto_size=auto_size, bar_mode=bar_mode, bg_color=bg_color,
	                   font_size=font_size, show_title=show_title, title=title)
	update_layout_axes(fig,
	                   axis_color=axis_color, axis_width=axis_width,
	                   grid_color=grid_color, grid_width=grid_width,
	                   label_color=label_color, label_size=label_size,
	                   range_to_zero_x=range_to_zero_x, range_to_zero_y=range_to_zero_y,
	                   range_to_zero_y2=range_to_zero_y2,
	                   show_grid_x=show_grid_x, show_grid_y=show_grid_y, show_grid_y2=show_grid_y2,
	                   show_spine=show_spine, show_zero_line=show_zero_line,
	                   tick_color=tick_color, tick_direction=tick_direction,
	                   tick_length=tick_length,
	                   tick_number_x=tick_number_x, tick_number_y=tick_number_y,
	                   tick_number_y2=tick_number_y2,
	                   tick_start_x=tick_start_x, tick_start_y=tick_start_y,
	                   tick_start_y2=tick_start_y2,
	                   tick_step_x=tick_step_x, tick_step_y=tick_step_y,
	                   tick_step_y2=tick_step_y2,
	                   tick_values_x=tick_values_x, tick_values_y=tick_values_y,
	                   tick_values_y2=tick_values_y2,
	                   title_x=title_x, title_y=title_y, title_y2=title_y2,
	                   zero_line_color=zero_line_color, zero_line_width=zero_line_width)
	update_layout_legend(fig, bg_color=legend_bg_color, x=legend_x, y=legend_y)
	update_layout_size(fig, width=width, height=height, margin=margin)


def update_layout_plot(fig, auto_size=True, bar_mode=None, bg_color=DEFAULT_BG_COLOR,
                       font_size=DEFAULT_FONT_SIZE, show_title=True, title=None):
	if is_matplot(fig):
		for ax in fig.axes:
			if not is_null(bg_color):
				ax.set_facecolor(bg_color)
			if not is_null(title) or not show_title:
				ax.set_title(title if show_title else None)
	elif is_plotly(fig):
		fig.update_annotations(font_size=font_size)
		bg_color = format_rgb_color(bg_color)
		if not is_null(title) or not show_title:
			fig.update_layout(title=web.b(title) if show_title else None)
		if not is_null(bar_mode):
			fig.update_layout(barmode=bar_mode)
		fig.update_layout(autosize=auto_size, paper_bgcolor=bg_color, plot_bgcolor=bg_color)


def update_layout_axes(fig,
                       axis_color='black', axis_width=DEFAULT_LINE_WIDTH,
                       grid_color='lightgray', grid_width=1,
                       label_color='black', label_size=None,
                       range_to_zero_x=False, range_to_zero_y=False, range_to_zero_y2=False,
                       scale_ratio_y=None, scale_ratio_y2=None,
                       show_grid_x=True, show_grid_y=True, show_grid_y2=True, show_spine=True,
                       show_zero_line=True,
                       tick_color='black', tick_direction=DEFAULT_TICK_DIRECTION,
                       tick_length=DEFAULT_TICK_LENGTH,
                       tick_number_x=None, tick_number_y=None, tick_number_y2=None,
                       tick_start_x=None, tick_start_y=None, tick_start_y2=None,
                       tick_step_x=None, tick_step_y=None, tick_step_y2=None,
                       tick_values_x=None, tick_values_y=None, tick_values_y2=None,
                       title_x=None, title_y=None, title_y2=None,
                       zero_line_color='darkgray', zero_line_width=DEFAULT_LINE_WIDTH):
	if is_matplot(fig):
		for ax in fig.axes:
			# Set the titles
			# - Horizontal axis
			if not is_null(title_x):
				ax.set_xlabel(to_string(title_x))
			# - Vertical axis
			if not is_null(title_y):
				ax.set_ylabel(to_string(title_y))
			# Set the spines
			for _, spine in ax.spines.items():
				spine.set_color(axis_color)
				spine.set_linewidth(axis_width)
				spine.set_visible(show_spine)
			# Set the grids
			# - Horizontal axis
			if show_grid_x:
				ax.grid(axis='x', b=True, color=grid_color, linestyle='-', linewidth=grid_width)
			else:
				ax.grid(axis='x', b=False)
			# - Vertical axis
			if show_grid_y:
				ax.grid(axis='y', b=True, color=grid_color, linestyle='-', linewidth=grid_width)
			else:
				ax.grid(axis='y', b=False)
			# Set the scale
			if not is_null(scale_ratio_y):
				ax.axes.set_aspect('equal')
			# Set the ticks
			ax.tick_params(color=tick_color,
			               direction='out' if tick_direction == 'outside' else 'in',
			               labelcolor=label_color, labelsize=label_size,
			               length=tick_length, width=grid_width)
			# - Horizontal axis
			if range_to_zero_x:
				ax.set_xlim(left=0)
			if not is_null(tick_number_x):
				ax.xaxis.set_major_locator(mticker.MaxNLocator(tick_number_x))
			elif not is_null(tick_start_x):
				ax.xaxis.set_major_locator(mticker.IndexLocator(base=tick_step_x,
				                                                offset=tick_start_x))
			elif not is_null(tick_step_x):
				ax.xaxis.set_major_locator(mticker.MultipleLocator(base=tick_step_x))
			elif not is_null(tick_values_x):
				ax.set_xticks(tick_values_x)
			# - Vertical axis
			if range_to_zero_y:
				ax.set_ylim(bottom=0)
			if not is_null(tick_number_y):
				ax.yaxis.set_major_locator(mticker.MaxNLocator(tick_number_y))
			elif not is_null(tick_start_y):
				ax.yaxis.set_major_locator(mticker.IndexLocator(base=tick_step_y,
				                                                offset=tick_start_y))
			elif not is_null(tick_step_y):
				ax.yaxis.set_major_locator(mticker.MultipleLocator(base=tick_step_y))
			elif not is_null(tick_values_y):
				ax.set_yticks(tick_values_y)
	elif is_plotly(fig):
		axis_color = format_rgb_color(axis_color)
		grid_color = format_rgb_color(grid_color)
		label_color = format_rgb_color(label_color)
		tick_color = format_rgb_color(tick_color)
		zero_line_color = format_rgb_color(zero_line_color)
		# Set the titles
		# - Horizontal axis
		if not is_null(title_x):
			fig.update_xaxes(title=dict(font_color=label_color, font_size=label_size,
			                            text=to_string(title_x)))
		# - Primary vertical axis
		if not is_null(title_y):
			update_axis_y(fig, yaxis=1,
			              title=dict(font_color=label_color, font_size=label_size,
			                         text=to_string(title_y)))
		# - Secondary vertical axis
		if not is_null(title_y2):
			update_axis_y(fig, yaxis=2,
			              title=dict(font_color=label_color, font_size=label_size,
			                         text=to_string(title_y2)))
		# Set the scale
		if not is_null(scale_ratio_y):
			update_axis_y(fig, yaxis=1,
			              scaleanchor='x', scaleratio=scale_ratio_y)
		if not is_null(scale_ratio_y2):
			update_axis_y(fig, yaxis=2,
			              scaleanchor='x', scaleratio=scale_ratio_y2)
		# Set the layout
		# - Horizontal axis
		fig.update_xaxes(
			# Set the spine
			showline=show_spine, linecolor=axis_color, linewidth=axis_width,
			# Set the grid
			showgrid=show_grid_x, gridcolor=grid_color, gridwidth=grid_width,
			# Set the range
			rangemode='tozero' if range_to_zero_x else None,
			# Set the ticks
			tickmode='array' if not is_null(tick_values_x)
			else 'linear' if not is_null(tick_start_x) or not is_null(tick_step_x)
			else 'auto', nticks=tick_number_x, tick0=tick_start_x, dtick=tick_step_x,
			tickvals=tick_values_x,
			tickcolor=tick_color, ticks=tick_direction, ticklen=tick_length,
			tickwidth=grid_width,
			# Set the zero line
			zeroline=show_zero_line, zerolinecolor=zero_line_color,
			zerolinewidth=zero_line_width)
		# - Primary vertical axis
		update_axis_y(fig, yaxis=1,
		              # Set the spine
		              showline=show_spine, linecolor=axis_color, linewidth=axis_width,
		              # Set the grid
		              showgrid=show_grid_y, gridcolor=grid_color, gridwidth=grid_width,
		              # Set the range
		              rangemode='tozero' if range_to_zero_y else None,
		              # Set the ticks
		              tickmode='array' if not is_null(tick_values_y)
		              else 'linear' if not is_null(tick_start_y) or not is_null(tick_step_y)
		              else 'auto', nticks=tick_number_y, tick0=tick_start_y, dtick=tick_step_y,
		              tickvals=tick_values_y,
		              tickcolor=tick_color, ticks=tick_direction, ticklen=tick_length,
		              tickwidth=grid_width,
		              # Set the zero line
		              zeroline=show_zero_line, zerolinecolor=zero_line_color,
		              zerolinewidth=zero_line_width)
		# - Secondary vertical axis
		update_axis_y(fig, yaxis=2,
		              # Set the spine
		              showline=show_spine, linecolor=axis_color, linewidth=axis_width,
		              # Set the grid
		              showgrid=show_grid_y2, gridcolor=grid_color, gridwidth=grid_width,
		              # Set the range
		              rangemode='tozero' if range_to_zero_y2 else None,
		              # Set the ticks
		              tickmode='array' if not is_null(tick_values_y2)
		              else 'linear' if not is_null(tick_start_y2) or not is_null(tick_step_y2)
		              else 'auto', nticks=tick_number_y2, tick0=tick_start_y2, dtick=tick_step_y2,
		              tickvals=tick_values_y2,
		              tickcolor=tick_color, ticks=tick_direction, ticklen=tick_length,
		              tickwidth=grid_width,
		              # Set the zero line
		              zeroline=show_zero_line, zerolinecolor=zero_line_color,
		              zerolinewidth=zero_line_width)


def update_axis_y(fig, yaxis=0, **kwargs):
	if is_multi_plot(fig):
		if yaxis == 2:
			fig.update_yaxes(secondary_y=True, **kwargs)
		else:
			fig.update_yaxes(secondary_y=False, **kwargs)
	else:
		if yaxis == 2:
			fig.update_layout(yaxis2=dict(**kwargs))
		else:
			fig.update_layout(yaxis=dict(**kwargs))


def update_layout_legend(fig, bg_color=DEFAULT_BG_COLOR, x=0.01, y=0.99):
	if is_matplot(fig):
		for ax in fig.axes:
			handles, _ = ax.get_legend_handles_labels()
			if not is_empty(handles):
				ax.legend(facecolor=bg_color, loc='upper left', bbox_to_anchor=(x, y))
	elif is_plotly(fig):
		bg_color = format_rgb_color(bg_color)
		fig.update_layout(legend=dict(xanchor='left', x=x,
		                              yanchor='top', y=y,
		                              bgcolor=bg_color))


def update_layout_size(fig, width=DEFAULT_WIDTH, height=DEFAULT_HEIGHT, margin=None):
	margin = get_margin(margin, fig=fig)
	if is_matplot(fig):
		fig.set_size_inches(width / 100, height / 100)
		fig.subplots_adjust(left=margin['l'], right=1 - margin['r'],
		                    bottom=margin['b'], top=1 - margin['t'])
	elif is_plotly(fig):
		margin = margin.copy()
		margin['l'] *= width
		margin['r'] *= width
		margin['b'] *= height
		margin['t'] *= height
		fig.update_layout(width=width, height=height, margin=margin)



#!/usr/bin/env python
####################################################################################################
# NAME
#    <NAME> - contain graphical utility functions for Web
#
# SYNOPSIS
#    <NAME>
#
# AUTHOR
#    Written by Florian Barras (florian@barras.io).
#
# COPYRIGHT
#    Copyright © 2013-2022 Florian Barras <https://barras.io>.
#    The MIT License (MIT) <https://opensource.org/licenses/MIT>.
####################################################################################################

import socket

import requests
from xhtml2pdf import pisa

from nutil.common import *

####################################################################################################
# WEB FUNCTIONS
####################################################################################################

__WEB_____________________________________________ = ''

# • WEB HTML #######################################################################################

__WEB_HTML________________________________________ = ''


def string_to_html(s):
	s = (s.replace('&', '&amp;')
	     .replace(' ', '&nbsp;')
	     .replace('\'', '&apos;')
	     .replace('"', '&quot;')
	     .replace('<', '&lt;')
	     .replace('>', '&gt;'))
	s = replace(s, '\r\n|\r|\n', '<br />')
	s = replace(s, '\t', '&nbsp;&nbsp;&nbsp;&nbsp;')
	return s


#########################

def html_to_pdf(path, html):
	'''Converts the specified HTML code to PDF.'''
	with open(path, mode='wb') as f:
		status = pisa.CreatePDF(html, dest=f)
		return status.err == 0


##################################################

def tag(name, value=None, attributes=None):
	if is_null(attributes):
		attributes = {}
	return collapse('<', name,
	                collapse([collapse(' ', k, '=', dquote(v)) for k, v in attributes.items()]),
	                collapse('>', value, '</', name, '>') if not is_empty(value) else ' />')


#########################

def a(value, attributes=None):
	return tag('a', value, attributes)


def abbr(value, attributes=None):
	return tag('abbr', value, attributes)


def address(value, attributes=None):
	return tag('address', value, attributes)


def area(value, attributes=None):
	return tag('area', value, attributes)


def article(value, attributes=None):
	return tag('article', value, attributes)


def aside(value, attributes=None):
	return tag('aside', value, attributes)


def audio(value, attributes=None):
	return tag('audio', value, attributes)


def b(value, attributes=None):
	return tag('b', value, attributes)


def base(value, attributes=None):
	return tag('base', value, attributes)


def bdi(value, attributes=None):
	return tag('bdi', value, attributes)


def bdo(value, attributes=None):
	return tag('bdo', value, attributes)


def blockquote(value, attributes=None):
	return tag('blockquote', value, attributes)


def body(value, attributes=None):
	return tag('body', value, attributes)


def br(attributes=None):
	return tag('br', None, attributes)


def button(value, attributes=None):
	return tag('button', value, attributes)


def canvas(value, attributes=None):
	return tag('canvas', value, attributes)


def caption(value, attributes=None):
	return tag('caption', value, attributes)


def cite(value, attributes=None):
	return tag('cite', value, attributes)


def code(value, attributes=None):
	return tag('code', value, attributes)


def col(value, attributes=None):
	return tag('col', value, attributes)


def colgroup(value, attributes=None):
	return tag('colgroup', value, attributes)


def data(value, attributes=None):
	return tag('data', value, attributes)


def datalist(value, attributes=None):
	return tag('datalist', value, attributes)


def dd(value, attributes=None):
	return tag('dd', value, attributes)


def delete(value, attributes=None):
	return tag('del', value, attributes)


def details(value, attributes=None):
	return tag('details', value, attributes)


def dfn(value, attributes=None):
	return tag('dfn', value, attributes)


def dialog(value, attributes=None):
	return tag('dialog', value, attributes)


def div(value, attributes=None):
	return tag('div', value, attributes)


def dl(value, attributes=None):
	return tag('dl', value, attributes)


def dt(value, attributes=None):
	return tag('dt', value, attributes)


def em(value, attributes=None):
	return tag('em', value, attributes)


def embed(value, attributes=None):
	return tag('embed', value, attributes)


def fieldset(value, attributes=None):
	return tag('fieldset', value, attributes)


def figcaption(value, attributes=None):
	return tag('figcaption', value, attributes)


def figure(value, attributes=None):
	return tag('figure', value, attributes)


def footer(value, attributes=None):
	return tag('footer', value, attributes)


def form(value, attributes=None):
	return tag('form', value, attributes)


def h(value, heading=1, attributes=None):
	return tag('h' + str(heading), value, attributes)


def head(value, attributes=None):
	return tag('head', value, attributes)


def header(value, attributes=None):
	return tag('header', value, attributes)


def hr(value, attributes=None):
	return tag('hr', value, attributes)


def html(value, attributes=None):
	return tag('html', value, attributes)


def i(value, attributes=None):
	return tag('i', value, attributes)


def iframe(value, attributes=None):
	return tag('iframe', value, attributes)


def img(value, attributes=None):
	return tag('img', value, attributes)


def input(value, attributes=None):
	return tag('input', value, attributes)


def insert(value, attributes=None):
	return tag('ins', value, attributes)


def kbd(value, attributes=None):
	return tag('kbd', value, attributes)


def label(value, attributes=None):
	return tag('label', value, attributes)


def legend(value, attributes=None):
	return tag('legend', value, attributes)


def li(value, attributes=None):
	return tag('li', value, attributes)


def link(value, attributes=None):
	return tag('link', value, attributes)


def main(value, attributes=None):
	return tag('main', value, attributes)


def map(value, attributes=None):
	return tag('map', value, attributes)


def mark(value, attributes=None):
	return tag('mark', value, attributes)


def meta(value, attributes=None):
	return tag('meta', value, attributes)


def meter(value, attributes=None):
	return tag('meter', value, attributes)


def nav(value, attributes=None):
	return tag('nav', value, attributes)


def noscript(value, attributes=None):
	return tag('noscript', value, attributes)


def object(value, attributes=None):
	return tag('object', value, attributes)


def ol(value, attributes=None):
	return tag('ol', value, attributes)


def optgroup(value, attributes=None):
	return tag('optgroup', value, attributes)


def option(value, attributes=None):
	return tag('option', value, attributes)


def output(value, attributes=None):
	return tag('output', value, attributes)


def p(value, attributes=None):
	return tag('p', value, attributes)


def param(value, attributes=None):
	return tag('param', value, attributes)


def picture(value, attributes=None):
	return tag('picture', value, attributes)


def pre(value, attributes=None):
	return tag('pre', value, attributes)


def progress(value, attributes=None):
	return tag('progress', value, attributes)


def q(value, attributes=None):
	return tag('q', value, attributes)


def rp(value, attributes=None):
	return tag('rp', value, attributes)


def rt(value, attributes=None):
	return tag('rt', value, attributes)


def ruby(value, attributes=None):
	return tag('ruby', value, attributes)


def s(value, attributes=None):
	return tag('s', value, attributes)


def samp(value, attributes=None):
	return tag('samp', value, attributes)


def script(value, attributes=None):
	return tag('script', value, attributes)


def section(value, attributes=None):
	return tag('section', value, attributes)


def select(value, attributes=None):
	return tag('select', value, attributes)


def small(value, attributes=None):
	return tag('small', value, attributes)


def source(value, attributes=None):
	return tag('source', value, attributes)


def span(value, attributes=None):
	return tag('span', value, attributes)


def strong(value, attributes=None):
	return tag('strong', value, attributes)


def style(value, attributes=None):
	return tag('style', value, attributes)


def sub(value, attributes=None):
	return tag('sub', value, attributes)


def summary(value, attributes=None):
	return tag('summary', value, attributes)


def sup(value, attributes=None):
	return tag('sup', value, attributes)


def svg(value, attributes=None):
	return tag('svg', value, attributes)


def table(value, attributes=None):
	return tag('table', value, attributes)


def tbody(value, attributes=None):
	return tag('tbody', value, attributes)


def td(value, attributes=None):
	return tag('td', value, attributes)


def template(value, attributes=None):
	return tag('template', value, attributes)


def textarea(value, attributes=None):
	return tag('textarea', value, attributes)


def tfoot(value, attributes=None):
	return tag('tfoot', value, attributes)


def th(value, attributes=None):
	return tag('th', value, attributes)


def thead(value, attributes=None):
	return tag('thead', value, attributes)


def time(value, attributes=None):
	return tag('time', value, attributes)


def title(value, attributes=None):
	return tag('title', value, attributes)


def tr(value, attributes=None):
	return tag('tr', value, attributes)


def track(value, attributes=None):
	return tag('track', value, attributes)


def u(value, attributes=None):
	return tag('u', value, attributes)


def ul(value, attributes=None):
	return tag('ul', value, attributes)


def var(value, attributes=None):
	return tag('var', value, attributes)


def video(value, attributes=None):
	return tag('video', value, attributes)


def wbr(value, attributes=None):
	return tag('wbr', value, attributes)


# • WEB NETWORK ####################################################################################

__WEB_NETWORK_____________________________________ = ''


def get_host_ip():
	'''Returns the IP of the host.'''
	return socket.gethostbyname(get_host_name())


def get_host_name():
	'''Returns the name of the host.'''
	return socket.gethostname()


# • WEB QUERY ######################################################################################

__WEB_QUERY_______________________________________ = ''


def download(url, dir=None):
	'''Downloads the file pointed by the specified URL and writes it to the specified directory.'''
	if is_null(dir):
		dir = get_dir('.')
	debug('Download the file', quote(url), 'to the directory', quote(dir))
	try:
		content = requests.get(url).content
		filename = get_filename(url).split('?')[0]
		return write_bytes(collapse(dir, '/', filename), content)
	except Exception as ex:
		error(ex)
		raise



#!/usr/bin/env python
####################################################################################################
# NAME
#    <NAME> - test the graphical utility functions
#
# SYNOPSIS
#    <NAME>
#
# AUTHOR
#    Written by Florian Barras (florian@barras.io).
#
# COPYRIGHT
#    Copyright © 2013-2022 Florian Barras <https://barras.io>.
#    The MIT License (MIT) <https://opensource.org/licenses/MIT>.
####################################################################################################

from ngui.chart import *
from ngui.image import *
from nutil.test import *

####################################################################################################
# GUI TEST CONSTANTS
####################################################################################################

__GUI_TEST_CONSTANTS______________________________ = ''

SIZE = 1000

####################################################################################################
# GUI TEST CLASSES
####################################################################################################

__GUI_TEST_CLASSES________________________________ = ''


class TestGui(Test):

	def test_chart(self):
		rgb_buffer = rotate_anti_90(generate_image(300, 200, 3))
		fig = px.imshow(rgb_buffer)

		self.assert_equals(sum(sum(sum(image_to_buffer(fig_to_jpg(fig))))), 367126580)
		self.assert_equals(sum(sum(sum(image_to_buffer(fig_to_png(fig))))), 524221335)
		self.assert_equals(sum(sum(sum(image_to_buffer(fig_to_webp(fig))))), 367221029)

		self.assert_equals(fig_to_jpg_html(fig)[-42:-10], 'FABRRRQAUUUUAFFFFABRRRQAUUUUAFFF')
		self.assert_equals(fig_to_png_html(fig)[-42:-10], 'AAAAAg0v8FMImIFolHcksAAAAASUVORK')
		self.assert_equals(fig_to_svg_html(fig)[-42:-10], 'g-xtitle"/><g class="g-ytitle"/>')
		self.assert_equals(fig_to_webp_html(fig)[-42:-10], '5z/7n/3H/uP/ef+8/95/5z/7n/3H/uv8')

	#####################

	def test_image(self):
		rgb = to_rgb(200, 100, 0, scale=False)
		hsv = rgb_to_hsv(rgb, scale=False)  # (0.083, 1.000, 0.784)
		self.assert_equals(hsv_to_rgb(hsv), rgb)

		self.assert_equals(get_complementary_color(rgb, scale=True), (0, 100, 200))

		self.assert_equals(get_base_color_name(rgb), 'y')
		self.assert_equals(get_tableau_color_name(rgb), 'tab:orange')
		self.assert_equals(get_css4_color_name(rgb), 'chocolate')
		self.assert_equals(get_xkcd_color_name(rgb), 'xkcd:browny orange')

		self.assert_equals(get_base_color_name(hsv, is_hsv=True), 'y')
		self.assert_equals(get_tableau_color_name(hsv, is_hsv=True), 'tab:olive')
		self.assert_equals(get_css4_color_name(hsv, is_hsv=True), 'darkgoldenrod')
		self.assert_equals(get_xkcd_color_name(hsv, is_hsv=True), 'xkcd:browny orange')

		self.assert_equals(format_rgb_color(rgb, alpha=None), 'rgb(200,100,0)')
		self.assert_equals(format_rgb_color(rgb), 'rgba(200,100,0,1)')

		self.assert_equals(format_hsv_color(hsv, alpha=None), 'hsv(30,100%,78%)')
		self.assert_equals(format_hsv_color(hsv), 'hsva(30,100%,78%,1)')

		self.assert_equals(format_rgb_color(map_to_color(100, normalize=True)),
		                   'rgba(0,104,55,1.0)')

		rgb_buffer = generate_image(300, 200, 3)
		hsv_buffer = cv2.cvtColor(rgb_buffer, cv2.COLOR_RGB2HSV)

		self.assert_equals(hsv_to_rgb(hsv_buffer, scale=False)[0],
		                   to_rgb(rgb_buffer, scale=False)[0], precision=1)
		self.assert_equals(rgb_to_hsv(rgb_buffer, scale=False)[0],
		                   to_hsv(hsv_buffer, scale=False)[0], precision=1)

		self.assert_equals(image_to_buffer(buffer_to_image(rgb_buffer, BMP_FORMAT))[0],
		                   rgb_buffer[0])

		self.assert_equals(buffer_to_html(rgb_buffer, BMP_FORMAT, rotate=True)[-42:-10],
		                   'eGSbtiKk/I3KxKINurU+/j0p12l32DAO')
		self.assert_equals(buffer_to_html(rgb_buffer, JPEG_FORMAT, rotate=True)[-42:-10],
		                   'ucPaXnVi+ZyX8GapwfuShaXLJ8zja97W')
		self.assert_equals(buffer_to_html(rgb_buffer, PNG_FORMAT, rotate=True)[-42:-10],
		                   'q21NUEUdmYyUW1RCNWQjkhAAAAAElFTk')
		self.assert_equals(buffer_to_html(rgb_buffer, TIFF_FORMAT, rotate=True)[-42:-10],
		                   'B4CAMArTMDAO9eAwA8igMAc7UDAAEAAQ')

		self.assert_equals(evaluate_colorfulness(rgb_buffer), 112.40, precision=2)
		self.assert_equals(evaluate_blurriness(rgb_buffer), 108379.74, precision=2)
		self.assert_equals(evaluate_brightness(rgb_buffer), 190.56, precision=2)

		# show_image(rgb_buffer)

	##############################################

	def evaluate_colorfulness(self, image):
		t = timeit.timeit(stmt=lambda: evaluate_colorfulness(image), number=TEST_COUNT)
		test('Evaluate the colorfulness of an image', 10 * TEST_COUNT, 'times in', round(t), '[s]')

	def evaluate_brightness(self, image):
		t = timeit.timeit(stmt=lambda: evaluate_brightness(image), number=TEST_COUNT)
		test('Evaluate the brightness of an image', 10 * TEST_COUNT, 'times in', round(t), '[s]')

	def evaluate_blurriness(self, image):
		t = timeit.timeit(stmt=lambda: evaluate_blurriness(image), number=TEST_COUNT)
		test('Evaluate the blurriness of an image', 10 * TEST_COUNT, 'times in', round(t), '[s]')


####################################################################################################
# GUI TEST MAIN
####################################################################################################

__GUI_TEST_MAIN___________________________________ = ''


def main():
	'''Tests the graphical utility library.'''
	unittest.main()


if __name__ == '__main__':
	main()



#!/usr/bin/env python
####################################################################################################
# NAME
#    <NAME> - contain graphical utility functions for images
#
# SYNOPSIS
#    <NAME>
#
# AUTHOR
#    Written by Florian Barras (florian@barras.io).
#
# COPYRIGHT
#    Copyright © 2013-2022 Florian Barras <https://barras.io>.
#    The MIT License (MIT) <https://opensource.org/licenses/MIT>.
####################################################################################################

import base64

from ngui.common import *
from nutil.math import *

####################################################################################################
# IMAGE CONSTANTS
####################################################################################################

__IMAGE_CONSTANTS_________________________________ = ''

DEFAULT_IMAGE_MODE = cv2.IMREAD_UNCHANGED

##################################################

BMP_FORMAT = 'bmp'  # bitmap
JPEG_FORMAT = 'jpg'  # Joint Photographic Experts Group
PNG_FORMAT = 'png'  # Portable Network Graphics
SVG_FORMAT = 'svg'  # Scalable Vector Graphics
TIFF_FORMAT = 'tiff'  # Tag Image File Format
WEBP_FORMAT = 'webp'  # WebP

####################################################################################################
# IMAGE FUNCTIONS
####################################################################################################

__IMAGE___________________________________________ = ''


def buffer_to_image(buffer, format, rotate=False):
	'''Converts the specified image buffer to an image of the specified format.'''
	if format == SVG_FORMAT:
		return buffer
	if rotate:
		buffer = rotate_anti_90(buffer)
	return cv2.imencode('.' + format, buffer)[1]


def buffer_to_html(buffer, format, encoding=DEFAULT_ENCODING, style=None,
                   rotate=False,
                   width=DEFAULT_WIDTH, height=DEFAULT_HEIGHT):
	'''Encodes the specified image buffer to Base64 and returns its HTML code.'''
	if is_empty(buffer):
		return ''
	if format == SVG_FORMAT:
		return buffer.decode(encoding)
	if rotate:
		width, height = height, width
	image = base64.b64encode(buffer_to_image(buffer, format, rotate=rotate)).decode(encoding)
	template = paste('<img width="{width}" height="{height}" src="data:image/{format};base64,{image}"',
	                 collapse('style="', style, '"') if not is_null(style) else '', '/>')
	return template.format(image=image, format=format, width=width, height=height)


#########################

def image_to_buffer(image, mode=DEFAULT_IMAGE_MODE):
	'''Converts the specified image to an image buffer.'''
	if image[1:4] == SVG_FORMAT.encode(DEFAULT_ENCODING):
		return image
	if is_string(image):
		image = read_bytes(image)
	return cv2.imdecode(np.frombuffer(image, dtype=SHORT_ELEMENT_TYPE), flags=mode)


def image_to_html(image, format, encoding=DEFAULT_ENCODING, mode=DEFAULT_IMAGE_MODE, style=None,
                  rotate=False,
                  width=DEFAULT_WIDTH, height=DEFAULT_HEIGHT):
	'''Converts the specified image to HTML.'''
	if format == SVG_FORMAT:
		return image.decode(encoding)
	return buffer_to_html(image_to_buffer(image, mode=mode), format, encoding=encoding, style=style,
	                      rotate=rotate,
	                      width=width, height=height)


##################################################

def generate_image(*shape):
	return create_random_short_array(255, *shape)


##################################################

def evaluate_colorfulness(buffer):
	'''Evaluates the colorfulness with the combination of the means and standard deviations of the
	color components of the specified RGB image buffer.'''
	# Split the color components of the RGB image buffer
	r, g, b = cv2.split(to_float(buffer))
	# Compute rg = R - G
	rg = cv2.absdiff(r, g)
	# Compute yb = 0.5 * (R + G) - B
	yb = cv2.absdiff(cv2.addWeighted(r, 0.5, g, 0.5, 0), b)
	# Compute the mean and standard deviation of rg and yb
	rg_mean, rg_std = cv2.meanStdDev(rg)
	yb_mean, yb_std = cv2.meanStdDev(yb)
	# Combine the means and standard deviations
	mean_root = np.sqrt(rg_mean ** 2 + yb_mean ** 2)
	std_root = np.sqrt(rg_std ** 2 + yb_std ** 2)
	# Derive the colorfulness metric
	return simplify(0.3 * mean_root + std_root)


def evaluate_blurriness(buffer):
	'''Evaluates the blurriness with the variance of the Laplacian of the specified image buffer.'''
	return cv2.Laplacian(buffer, cv2.CV_64F).var()


def evaluate_brightness(buffer):
	'''Evaluates the blurriness with the mean of the value (brightness) of the HSV representation of
	the specified image buffer.'''
	_, _, v = rgb_to_hsv(buffer)
	return mean(mean(v))


#########################

def resize_image(buffer,
                 width=DEFAULT_WIDTH, height=DEFAULT_HEIGHT):
	return cv2.resize(buffer, (width, height))


#########################

def rotate_anti_90(buffer):
	buffer = cv2.transpose(buffer)
	buffer = cv2.flip(buffer, 0)
	return buffer


def rotate_anti_270(buffer):
	buffer = cv2.transpose(buffer)
	buffer = cv2.flip(buffer, 1)
	return buffer


def rotate_by(buffer, angle, center=None, scale=1):
	h, w = buffer.shape[:2]
	if is_null(center):
		center = (w / 2, h / 2)
	m = cv2.getRotationMatrix2D(center, angle, scale)
	return cv2.warpAffine(buffer, m, (w, h))


#########################

def show_image(buffer, title='Image'):
	cv2.imshow(title, buffer)
	cv2.waitKey(0)
	cv2.destroyAllWindows()


##################################################

def write_image(path, buffer):
	return cv2.imwrite(path, buffer)



#!/usr/bin/env python
####################################################################################################
# NAME
#    <NAME> - contain common graphical utility functions
#
# SYNOPSIS
#    <NAME>
#
# AUTHOR
#    Written by Florian Barras (florian@barras.io).
#
# COPYRIGHT
#    Copyright © 2013-2022 Florian Barras <https://barras.io>.
#    The MIT License (MIT) <https://opensource.org/licenses/MIT>.
####################################################################################################

import cv2
import matplotlib.cm as mcm
import matplotlib.colors as mcolors

from nmath import normal
from nutil.math import *

####################################################################################################
# GUI COMMON CONSTANTS
####################################################################################################

__GUI_COMMON_CONSTANTS____________________________ = ''

# The default scale
DEFAULT_SCALE = 1  # the higher, the better quality

# The default width
DEFAULT_WIDTH = 660

# The default height
DEFAULT_HEIGHT = 933

# The default margin (left, right, bottom and top) defined by the ratio to the width or height
DEFAULT_MARGIN = dict(l=0, r=0, b=0, t=0)
DEFAULT_MARGIN_WITH_TITLE = dict(l=0.05, r=0.05, b=0.05, t=0.05)

#########################

# The default font size
DEFAULT_FONT_SIZE = 12

# The default line width
DEFAULT_LINE_WIDTH = 2

# The default marker size
DEFAULT_MARKER_SIZE = 4

# • GUI COLOR ######################################################################################

__GUI_COLOR_CONSTANTS_____________________________ = ''

TRANSPARENT = (0, 0, 0, 0)

##################################################

# The default colors
DEFAULT_COLORS = [
	'#1F77B4',  # muted blue
	'#FF7F0E',  # safety orange
	'#2CA02C',  # cooked asparagus green
	'#D62728',  # brick red
	'#9467BD',  # muted purple
	'#8C564B',  # chestnut brown
	'#E377C2',  # raspberry yogurt pink
	'#7F7F7F',  # middle gray
	'#BCBD22',  # curry yellow-green
	'#17BECF'  # blue-teal
]

# The default colors iterator
DEFAULT_COLORS_ITERATOR = get_iterator(DEFAULT_COLORS, cycle=True)

# The default background color
DEFAULT_BG_COLOR = TRANSPARENT


##################################################

def to_rgb(*args, r=0, g=0, b=0, alpha=None, scale=None):
	color = forward_element(*args)
	if is_collection(color) or is_tuple(color):
		if is_array(color):
			r, g, b = cv2.split(to_float(color))
		elif len(color) == 3:
			r, g, b = color
		elif len(color) == 4:
			r, g, b, alpha = color
	elif is_string(color):
		if 'rgb' in color:
			color = to_float(extract(color, '[0-9\.]+'))
			if len(color) == 3:
				r, g, b = color
			elif len(color) == 4:
				r, g, b, alpha = color
		else:
			if not is_null(alpha):
				r, g, b, alpha = mcolors.to_rgba(color, alpha=alpha)
			else:
				r, g, b = mcolors.to_rgb(color)
	if not is_null(scale):
		r, g, b = unscale_color(r, g, b)
		if scale:
			r, g, b = scale_color(r, g, b)
	if not is_null(alpha):
		return r, g, b, alpha
	return r, g, b


#########################

def to_hsv(*args, h=0, s=0, v=0, alpha=None, scale=None):
	color = forward_element(*args)
	if is_collection(color) or is_tuple(color):
		if is_array(color):
			h, s, v = cv2.split(to_float(color))
		elif len(color) == 3:
			h, s, v = color
		elif len(color) == 4:
			h, s, v, alpha = color
	if not is_null(scale):
		h, s, v = unscale_color(h, s, v)
		if scale:
			h, s, v = scale_color(h, s, v)
	if not is_null(alpha):
		return h, s, v, alpha
	return h, s, v


#########################

def rgb_to_hsv(*args, r=0, g=0, b=0, alpha=None, scale=None):
	color = forward_element(*args)
	if is_array(color):
		image = cv2.cvtColor(color, cv2.COLOR_RGB2HSV)
		h, s, v = cv2.split(to_float(image))
	else:
		r, g, b = to_rgb(*args, r=r, g=g, b=b)
		h, s, v = mcolors.rgb_to_hsv((r, g, b))
	if not is_null(scale):
		h, s, v = unscale_color(h, s, v)
		if scale:
			h, s, v = scale_color(h, s, v)
	if not is_null(alpha):
		return h, s, v, alpha
	return h, s, v


def hsv_to_rgb(*args, h=0, s=0, v=0, alpha=None, scale=None):
	color = forward_element(*args)
	if is_array(color):
		image = cv2.cvtColor(color, cv2.COLOR_HSV2RGB)
		r, g, b = cv2.split(to_float(image))
	else:
		h, s, v = to_hsv(*args, h=h, s=s, v=v)
		r, g, b = mcolors.hsv_to_rgb((h, s, v))
	if not is_null(scale):
		r, g, b = unscale_color(r, g, b)
		if scale:
			r, g, b = scale_color(r, g, b)
	if not is_null(alpha):
		return r, g, b, alpha
	return r, g, b


#########################

def is_scaled_color(x, y, z):
	return maximum(maximum(x)) > 1 or maximum(maximum(y)) > 1 or maximum(maximum(z)) > 1


def is_unscaled_color(x, y, z):
	return maximum(maximum(x)) <= 1 and maximum(maximum(y)) <= 1 and maximum(maximum(z)) <= 1


#########################

def scale_color(x, y, z):
	if is_unscaled_color(x, y, z):
		x = round_to_int(x * 255)
		y = round_to_int(y * 255)
		z = round_to_int(z * 255)
	return x, y, z


def unscale_color(x, y, z):
	if is_scaled_color(x, y, z):
		x /= 255
		y /= 255
		z /= 255
	return x, y, z


#########################

BASE_COLOR_NAMES = [name for name, _ in mcolors.BASE_COLORS.items()]
BASE_COLOR_RGB_CODES = to_array([to_rgb(color) for _, color in mcolors.BASE_COLORS.items()])
BASE_COLOR_HSV_CODES = to_array([rgb_to_hsv(color) for _, color in mcolors.BASE_COLORS.items()])

TABLEAU_COLOR_NAMES = [name for name, _ in mcolors.TABLEAU_COLORS.items()]
TABLEAU_COLOR_RGB_CODES = to_array([to_rgb(color) for _, color in mcolors.TABLEAU_COLORS.items()])
TABLEAU_COLOR_HSV_CODES = to_array([rgb_to_hsv(color) for _, color in mcolors.TABLEAU_COLORS.items()])

CSS4_COLOR_NAMES = [name for name, _ in mcolors.CSS4_COLORS.items()]
CSS4_COLOR_RGB_CODES = to_array([to_rgb(color) for _, color in mcolors.CSS4_COLORS.items()])
CSS4_COLOR_HSV_CODES = to_array([rgb_to_hsv(color) for _, color in mcolors.CSS4_COLORS.items()])

XKCD_COLOR_NAMES = [name for name, _ in mcolors.XKCD_COLORS.items()]
XKCD_COLOR_RGB_CODES = to_array([to_rgb(color) for _, color in mcolors.XKCD_COLORS.items()])
XKCD_COLOR_HSV_CODES = to_array([rgb_to_hsv(color) for _, color in mcolors.XKCD_COLORS.items()])

RAINBOW_SCALE = mcm.get_cmap(name='rainbow')
RYG_SCALE = mcm.get_cmap(name='RdYlGn')

####################################################################################################
# GUI COMMON FUNCTIONS
####################################################################################################

__GUI_COMMON______________________________________ = ''

# • GUI COLOR ######################################################################################

__GUI_COLOR_______________________________________ = ''


def get_alternate_colors(n, row_odd_color='white', row_even_color='lightgray'):
	return ceil(n / 2) * [row_odd_color, row_even_color]


def get_complementary_color(*args, r=0, g=0, b=0, scale=None):
	r, g, b = to_rgb(*args, r=r, g=g, b=b, scale=scale)
	return tuple(minimum(r, g, b) + maximum(r, g, b) - c for c in (r, g, b))


def get_RYG(brightness='8'):
	colors = ('#E.0.0.', '#E..00.', '#E.E.0.', '#.0E.00', '#0...0.')
	return [color.replace('.', brightness) for color in colors]


#########################

def get_random_base_color_name():
	return random.choice(BASE_COLOR_NAMES)


def get_random_tableau_color_name():
	return random.choice(TABLEAU_COLOR_NAMES)


def get_random_css4_color_name():
	return random.choice(CSS4_COLOR_NAMES)


def get_random_xkcd_color_name():
	return random.choice(XKCD_COLOR_NAMES)


#########################

def get_color_name(code, names, codes):
	return names[min_distance_index(code, codes)]


def get_base_color_name(code, is_hsv=False):
	return get_color_name(unscale_color(*code), BASE_COLOR_NAMES,
	                      BASE_COLOR_HSV_CODES if is_hsv else BASE_COLOR_RGB_CODES)


def get_tableau_color_name(code, is_hsv=False):
	return get_color_name(unscale_color(*code), TABLEAU_COLOR_NAMES,
	                      TABLEAU_COLOR_HSV_CODES if is_hsv else TABLEAU_COLOR_RGB_CODES)


def get_css4_color_name(code, is_hsv=False):
	return get_color_name(unscale_color(*code), CSS4_COLOR_NAMES,
	                      CSS4_COLOR_HSV_CODES if is_hsv else CSS4_COLOR_RGB_CODES)


def get_xkcd_color_name(code, is_hsv=False):
	return get_color_name(unscale_color(*code), XKCD_COLOR_NAMES,
	                      XKCD_COLOR_HSV_CODES if is_hsv else XKCD_COLOR_RGB_CODES)


##################################################

def format_rgb_color(*args, r=0, g=0, b=0, alpha=1):
	'''Formats the specified RGB color (with alpha if it is not null).'''
	prefix = 'rgba' if not is_null(alpha) else 'rgb'
	return prefix + par(collist(*to_rgb(*args, r=r, g=g, b=b, alpha=alpha, scale=True)))


def format_hsv_color(*args, h=0, s=0, v=0, alpha=1):
	'''Formats the specified HSV color (with alpha if it is not null).'''
	prefix = 'hsva' if not is_null(alpha) else 'hsv'
	h, s, v = to_hsv(*args, h=h, s=s, v=v, scale=False)
	hsv = (round_to_int(h * 360), format_percent(s, 0), format_percent(v, 0))
	return prefix + par(collist(*hsv, alpha) if not is_null(alpha) else collist(*hsv))


#########################

def map_to_color(value, color_scale=RYG_SCALE, normalize=False):
	'''Formats the specified value to an RGBA color using the specified alpha and color scale.'''
	if normalize:
		value = float(normal.cdf(value))
	return color_scale(value)
